import enum
import logging
import struct

from ..wire.exceptions import PacketInvalidData
from ..wire.read import (
    read_string,
    read_uint8,
    read_uint16,
    read_uint32,
)
from ..wire.tcp import TCPProtocol
from ..wire.write import (
    SEND_TCP_COMPAT_MTU,
    write_init,
    write_presend,
    write_string,
    write_uint8,
    write_uint32,
)

log = logging.getLogger(__name__)


# Copy from OpenTTD/src/network/core/tcp_content.h
class PacketContentType(enum.IntEnum):
    PACKET_CONTENT_CLIENT_INFO_LIST = 0
    PACKET_CONTENT_CLIENT_INFO_ID = 1
    PACKET_CONTENT_CLIENT_INFO_EXTID = 2
    PACKET_CONTENT_CLIENT_INFO_EXTID_MD5 = 3
    PACKET_CONTENT_SERVER_INFO = 4
    PACKET_CONTENT_CLIENT_CONTENT = 5
    PACKET_CONTENT_SERVER_CONTENT = 6
    PACKET_CONTENT_END = 7


class ContentType(enum.IntEnum):
    CONTENT_TYPE_BASE_GRAPHICS = 1
    CONTENT_TYPE_NEWGRF = 2
    CONTENT_TYPE_AI = 3
    CONTENT_TYPE_AI_LIBRARY = 4
    CONTENT_TYPE_SCENARIO = 5
    CONTENT_TYPE_HEIGHTMAP = 6
    CONTENT_TYPE_BASE_SOUNDS = 7
    CONTENT_TYPE_BASE_MUSIC = 8
    CONTENT_TYPE_GAME = 9
    CONTENT_TYPE_GAME_LIBRARY = 10
    CONTENT_TYPE_END = 11


class ContentInfo:
    def __init__(self, content_id=None, content_type=None, unique_id=None, md5sum=None):
        super().__init__()

        self.content_id = content_id
        self.content_type = content_type
        self.unique_id = unique_id
        self.md5sum = md5sum

    def __repr__(self):
        return (
            f"ContentInfo(content_id={self.content_id!r},"
            f"content_type={self.content_type!r}, "
            f"unique_id={self.unique_id!r}, "
            f"md5sum={self.md5sum!r})"
        )


class ContentProtocol(TCPProtocol):
    PacketType = PacketContentType
    PACKET_END = PacketContentType.PACKET_CONTENT_END

    @staticmethod
    def receive_PACKET_CONTENT_CLIENT_INFO_LIST(source, data):
        content_type, data = read_uint8(data)
        openttd_version, data = read_uint32(data)

        # Since OpenTTD 12.0 we extended this packet to include multiple
        # branches and their versions, so patchpacks can filter the list
        # based on their version. This is indicated by sending an
        # openttd_version that is UINT32_MAX.
        branch_versions = {}
        if openttd_version == 0xFFFFFFFF:
            count, data = read_uint8(data)
            for _ in range(count):
                branch, data = read_string(data)
                version, data = read_string(data)
                branch_versions[branch] = version

        if content_type >= ContentType.CONTENT_TYPE_END:
            raise PacketInvalidData("invalid ContentType", content_type)

        content_type = ContentType(content_type)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected; remaining: ", len(data))

        return {"content_type": content_type, "openttd_version": openttd_version, "branch_versions": branch_versions}

    @staticmethod
    def _receive_client_info(data, count, has_content_id=False, has_content_type_and_unique_id=False, has_md5sum=False):
        content_infos = []
        for _ in range(count):
            content_info = {}

            if has_content_id:
                content_id, data = read_uint32(data)
                content_info["content_id"] = content_id

            if has_content_type_and_unique_id:
                content_type, data = read_uint8(data)
                if content_type >= ContentType.CONTENT_TYPE_END:
                    raise PacketInvalidData("invalid ContentType", content_type)
                content_type = ContentType(content_type)
                content_info["content_type"] = content_type

                unique_id, data = read_uint32(data)
                if content_type == ContentType.CONTENT_TYPE_NEWGRF:
                    # OpenTTD client sends NewGRFs byte-swapped for some reason.
                    # So we swap it back here, as nobody needs to know the
                    # protocol is making a boo-boo.
                    content_info["unique_id"] = unique_id.to_bytes(4, "big")
                elif content_type in (ContentType.CONTENT_TYPE_SCENARIO, ContentType.CONTENT_TYPE_HEIGHTMAP):
                    # We store Scenarios / Heightmaps byte-swapped (to what OpenTTD expects).
                    # This is because otherwise folders are named 01000000, 02000000, which
                    # makes sorting a bit odd, and in general just difficult to read.
                    content_info["unique_id"] = unique_id.to_bytes(4, "big")
                else:
                    content_info["unique_id"] = unique_id.to_bytes(4, "little")

            if has_md5sum:
                md5sum = bytearray()
                for _ in range(16):
                    md5sum_snippet, data = read_uint8(data)
                    md5sum.append(md5sum_snippet)
                md5sum = bytes(md5sum)
                content_info["md5sum"] = md5sum

            content_infos.append(ContentInfo(**content_info))

        return content_infos, data

    @classmethod
    def receive_PACKET_CONTENT_CLIENT_INFO_ID(cls, source, data):
        count, data = read_uint16(data)

        content_infos, data = cls._receive_client_info(data, count, has_content_id=True)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected; remaining: ", len(data))

        return {"content_infos": content_infos}

    @classmethod
    def receive_PACKET_CONTENT_CLIENT_INFO_EXTID(cls, source, data):
        count, data = read_uint8(data)

        content_infos, data = cls._receive_client_info(data, count, has_content_type_and_unique_id=True)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected; remaining: ", len(data))

        return {"content_infos": content_infos}

    @classmethod
    def receive_PACKET_CONTENT_CLIENT_INFO_EXTID_MD5(cls, source, data):
        count, data = read_uint8(data)

        content_infos, data = cls._receive_client_info(
            data, count, has_content_type_and_unique_id=True, has_md5sum=True
        )

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected; remaining: ", len(data))

        return {"content_infos": content_infos}

    @classmethod
    def receive_PACKET_CONTENT_CLIENT_CONTENT(cls, source, data):
        count, data = read_uint16(data)

        content_infos, data = cls._receive_client_info(data, count, has_content_id=True)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected; remaining: ", len(data))

        return {"content_infos": content_infos}

    async def send_PACKET_CONTENT_SERVER_INFO(
        self, content_type, content_id, filesize, name, version, url, description, unique_id, md5sum, dependencies, tags
    ):
        data = write_init(PacketContentType.PACKET_CONTENT_SERVER_INFO)

        write_uint8(data, content_type.value)
        write_uint32(data, content_id)

        write_uint32(data, filesize)
        write_string(data, name)
        write_string(data, version)
        write_string(data, url)
        write_string(data, description)

        if content_type == ContentType.CONTENT_TYPE_NEWGRF:
            # OpenTTD client sends NewGRFs byte-swapped for some reason.
            # So we swap it back here, as nobody needs to know the
            # protocol is making a boo-boo.
            write_uint32(data, struct.unpack(">I", unique_id)[0])
        elif content_type in (ContentType.CONTENT_TYPE_SCENARIO, ContentType.CONTENT_TYPE_HEIGHTMAP):
            # We store Scenarios / Heightmaps byte-swapped (to what OpenTTD expects).
            # This is because otherwise folders are named 01000000, 02000000, which
            # makes sorting a bit odd, and in general just difficult to read.
            write_uint32(data, struct.unpack(">I", unique_id)[0])
        else:
            write_uint32(data, struct.unpack("<I", unique_id)[0])

        for i in range(16):
            write_uint8(data, md5sum[i])

        write_uint8(data, len(dependencies))
        for dependency in dependencies:
            write_uint32(data, dependency)

        write_uint8(data, len(tags))
        for tag in tags:
            write_string(data, tag)

        write_presend(data, SEND_TCP_COMPAT_MTU)
        await self.send_packet(data)

    async def send_PACKET_CONTENT_SERVER_CONTENT(self, content_type, content_id, filesize, filename, stream):
        # First, send a packet to tell the client it will be receiving a file
        data = write_init(PacketContentType.PACKET_CONTENT_SERVER_CONTENT)

        write_uint8(data, content_type.value)
        write_uint32(data, content_id)

        write_uint32(data, filesize)
        write_string(data, filename)

        write_presend(data, SEND_TCP_COMPAT_MTU)
        await self.send_packet(data)

        # Next, send the content of the file over
        while not stream.eof():
            data = write_init(PacketContentType.PACKET_CONTENT_SERVER_CONTENT)
            data += stream.read(SEND_TCP_COMPAT_MTU - 3)
            write_presend(data, SEND_TCP_COMPAT_MTU)
            await self.send_packet(data)

        data = write_init(PacketContentType.PACKET_CONTENT_SERVER_CONTENT)
        write_presend(data, SEND_TCP_COMPAT_MTU)
        await self.send_packet(data)



import os
import shlex
import subprocess


def _get_git_version(path):
    # Check if we are a git clone.
    if not os.path.isdir(os.path.join(path, ".git")):
        return None

    # Base the version off the "git describe --tags".
    command = shlex.split(f"git -C {path} describe --tags --long")
    try:
        describe = subprocess.check_output(command, universal_newlines=True).strip()
    except subprocess.CalledProcessError:
        # Most likely this was a clone without tags
        return None
    version, commit_since, git_hash = describe.rsplit("-", 3)

    # If we are not a release, we are a "post"; locally we can also be a
    # "dev", but as it is locally, it doesn't really matter. The hash will
    # change anyway, showing it is a different version.
    if commit_since != "0":
        return f"{version}.post{commit_since}+{git_hash}"

    return version


def get_version():
    path = os.path.dirname(os.path.realpath(__file__))
    version_file = os.path.join(path, "VERSION")

    # Load the version from disk if possible.
    if os.path.isfile(version_file):
        with open(version_file, "r") as f:
            disk_version = f.read()
    else:
        disk_version = None

    # Check if we have a git version.
    git_version = _get_git_version(path)

    # If we don't have a git version, return the disk version.
    if not git_version:
        return disk_version

    # IF the git version and disk version differ, write a new disk version.
    if git_version != disk_version:
        with open(version_file, "w") as f:
            f.write(git_version)

    return git_version



import asyncio
import enum
import pytest

from .exceptions import (
    PacketInvalidData,
    PacketInvalidSize,
    PacketInvalidType,
)
from .source import Source
from .tcp import TCPProtocol
from .read import read_uint8


class OpenTTDTestType(enum.Enum):
    PACKET_ONE = 0
    PACKET_TWO = 1
    PACKET_THREE = 2
    PACKET_END = 3


class OpenTTDProtocolTest(TCPProtocol):
    PacketType = OpenTTDTestType
    PACKET_END = PacketType.PACKET_END.value

    def receive_PACKET_ONE(self, source, data):
        assert data == b""
        return {}

    def receive_PACKET_TWO(self, source, data):
        value, data = read_uint8(data)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected; remaining: ", len(data))

        return {"value": value}


class FakeTransport:
    def close(self):
        pass


@pytest.mark.parametrize(
    "proxy_protocol, data, result, ip, port",
    [
        (False, b"\x03\x00\x00", b"\x03\x00\x00", "127.0.0.2", 54321),
        (True, b"PROXY TCP4 127.0.0.1 127.0.0.1 12345 12121\r\n\x03\x00\x00", b"\x03\x00\x00", "127.0.0.1", 12345),
        (True, b"\x03\x00\x00", b"\x03\x00\x00", "127.0.0.2", 54321),
    ],
)
@pytest.mark.asyncio
async def test_detect_source_ip_port(proxy_protocol, data, result, ip, port):
    test = OpenTTDProtocolTest(None)
    test.task.cancel()
    test.source = Source(test, None, "127.0.0.2", 54321)
    test.proxy_protocol = proxy_protocol

    assert test._detect_source_ip_port(memoryview(data)) == result
    assert str(test.source.ip) == ip
    assert test.source.port == port


@pytest.mark.parametrize(
    "data, data_left, result",
    [
        (b"\x03\x00\x00", b"", b"\x03\x00\x00"),
        (b"\x03\x00\x00\x03", b"\x03", b"\x03\x00\x00"),
        (b"\x05\x00\x00", b"\x05\x00\x00", b""),
    ],
)
@pytest.mark.asyncio
async def test_data_received(data, data_left, result):
    test = OpenTTDProtocolTest(None)
    test.task.cancel()

    test.data_received(memoryview(data))
    assert test._data == data_left
    try:
        assert test._queue.get_nowait() == result
    except asyncio.queues.QueueEmpty:
        assert b"" == result


@pytest.mark.parametrize(
    "data, result",
    [
        (b"\x03\x00\x00", (OpenTTDTestType.PACKET_ONE, {})),
        (b"\x04\x00\x01\x02", (OpenTTDTestType.PACKET_TWO, {"value": 2})),
    ],
)
@pytest.mark.asyncio
async def test_receive_packet(data, result):
    test = OpenTTDProtocolTest(None)
    test.task.cancel()

    assert test.receive_packet(None, memoryview(data)) == result


@pytest.mark.parametrize(
    "data, failure",
    [
        (b"\x04\x00\x00", PacketInvalidSize),
        (b"\x03\x00\x02", PacketInvalidType),
        (b"\x03\x00\x03", PacketInvalidType),
        (b"\x05\x00\x01\x02\x00", PacketInvalidData),
    ],
)
@pytest.mark.asyncio
async def test_receive_packet_failure(data, failure):
    test = OpenTTDProtocolTest(None)
    test.task.cancel()

    with pytest.raises(failure):
        test.receive_packet(None, memoryview(data))


@pytest.mark.parametrize(
    "data, result",
    [
        (b"\x03\x00\x00", {}),
        (b"\x04\x00\x01\x02", {"value": 2}),
    ],
)
@pytest.mark.asyncio
async def test_process_queue(data, result):
    seen_packet = [False]

    class Callback:
        async def receive_PACKET_ONE(source, **kwargs):
            assert kwargs == result
            seen_packet[0] = True

        async def receive_PACKET_TWO(source, **kwargs):
            assert kwargs == result
            seen_packet[0] = True

    test = OpenTTDProtocolTest(Callback)
    test.task.cancel()

    test.source = Source(test, None, "127.0.0.1", 12345)
    test.transport = FakeTransport()

    test._queue.put_nowait(memoryview(data))
    test._queue.put_nowait(memoryview(b"\x04\x00\x00"))  # Force an exception
    await test._process_queue()
    assert seen_packet[0] is True



import enum
import logging

from ..wire.exceptions import PacketInvalidData
from ..wire.read import (
    read_bytes,
    read_string,
    read_uint8,
    read_uint16,
    read_uint32,
)
from ..wire.tcp import TCPProtocol
from ..wire.write import (
    SEND_TCP_MTU,
    write_init,
    write_presend,
)

log = logging.getLogger(__name__)


# The minimum starting year on the original TTD.
ORIGINAL_BASE_YEAR = 1920
# In GameInfo version 3 the date was changed to be counted from the year zero.
# This offset is added to version 2 and 1 to have the date the same for all
# versions. It is the amount of days from year 0 to 1920.
DAYS_TILL_ORIGINAL_BASE_YEAR = (
    365 * ORIGINAL_BASE_YEAR + ORIGINAL_BASE_YEAR // 4 - ORIGINAL_BASE_YEAR // 100 + ORIGINAL_BASE_YEAR // 400
)


class PacketGameType(enum.IntEnum):
    # TODO -- Packets 0 .. 5 are not implemented yet. Pull Requests are welcome.
    PACKET_SERVER_GAME_INFO = 6
    PACKET_CLIENT_GAME_INFO = 7
    # TODO -- Packets 8 .. 39 are not implemented yet. Pull Requests are welcome.
    PACKET_SERVER_SHUTDOWN = 40
    # TODO -- Packets 41 .. 43 are not implemented yet. Pull Requests are welcome.
    PACKET_END = 44


class NewGRFSerializationType(enum.IntEnum):
    NST_GRFID_MD5 = 0
    NST_GRFID_MD5_NAME = 1
    NST_LOOKUP_ID = 2
    NST_END = 3
    # NST_CONVERSION_GRFID_MD5 is an internal value, assigned for those servers that didn't send this value yet.
    NST_CONVERSION_GRFID_MD5 = 4


class GameProtocol(TCPProtocol):
    PacketType = PacketGameType
    PACKET_END = PacketGameType.PACKET_END

    @staticmethod
    def receive_PACKET_SERVER_GAME_INFO(source, data):
        game_info_version, data = read_uint8(data)

        if game_info_version < 1 or game_info_version > 6:
            raise PacketInvalidData("unknown game info version: ", game_info_version)

        if game_info_version >= 6:
            newgrf_serialization_type, data = read_uint8(data)

            if newgrf_serialization_type >= NewGRFSerializationType.NST_END:
                raise PacketInvalidData("invalid NewGRFSerializationType", newgrf_serialization_type)

            newgrf_serialization_type = NewGRFSerializationType(newgrf_serialization_type)
            if newgrf_serialization_type == NewGRFSerializationType.NST_LOOKUP_ID:
                raise PacketInvalidData("NewGRF serialization type cannot be NST_LOOKUP_ID")
        else:
            newgrf_serialization_type = NewGRFSerializationType.NST_CONVERSION_GRFID_MD5

        if game_info_version >= 5:
            gamescript_version, data = read_uint32(data)
            gamescript_name, data = read_string(data)
        else:
            gamescript_version = None
            gamescript_name = None

        if game_info_version >= 4:
            newgrf_count, data = read_uint8(data)

            newgrfs = []
            for _ in range(newgrf_count):
                grfid, data = read_uint32(data)
                md5sum, data = read_bytes(data, 16)

                if newgrf_serialization_type == NewGRFSerializationType.NST_GRFID_MD5_NAME:
                    name, data = read_string(data)
                else:
                    name = None

                newgrfs.append({"grfid": grfid, "md5sum": md5sum.hex(), "name": name})
        else:
            newgrfs = None

        if game_info_version >= 3:
            game_date, data = read_uint32(data)
            start_date, data = read_uint32(data)

        if game_info_version >= 2:
            companies_max, data = read_uint8(data)
            companies_on, data = read_uint8(data)
            spectators_max, data = read_uint8(data)
        else:
            companies_max = None
            companies_on = None
            spectators_max = None

        if game_info_version >= 1:
            name, data = read_string(data)
            openttd_version, data = read_string(data)
            if game_info_version < 6:
                _, data = read_uint8(data)  # Unused, used to be server-lang
            use_password, data = read_uint8(data)

            clients_max, data = read_uint8(data)
            clients_on, data = read_uint8(data)
            spectators_on, data = read_uint8(data)

            if game_info_version < 3:
                game_date, data = read_uint16(data)
                game_date += DAYS_TILL_ORIGINAL_BASE_YEAR
                start_date, data = read_uint16(data)
                start_date += DAYS_TILL_ORIGINAL_BASE_YEAR

            if game_info_version < 6:
                _, data = read_string(data)  # Unused, used to be map-name
            map_width, data = read_uint16(data)
            map_height, data = read_uint16(data)
            map_type, data = read_uint8(data)

            is_dedicated, data = read_uint8(data)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected in SERVER_GAME_INFO; remaining: ", len(data))

        return {
            "newgrfs": newgrfs,
            "game_date": game_date,
            "start_date": start_date,
            "companies_max": companies_max,
            "companies_on": companies_on,
            "clients_max": clients_max,
            "clients_on": clients_on,
            "spectators_max": spectators_max,
            "spectators_on": spectators_on,
            "name": name,
            "openttd_version": openttd_version,
            "use_password": use_password,
            "is_dedicated": is_dedicated,
            "map_width": map_width,
            "map_height": map_height,
            "map_type": map_type,
            "gamescript_version": gamescript_version,
            "gamescript_name": gamescript_name,
        }

    @staticmethod
    def receive_PACKET_SERVER_SHUTDOWN(source, data):
        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected in SERVER_SHUTDOWN; remaining: ", len(data))

        return {}

    async def send_PACKET_CLIENT_GAME_INFO(self):
        data = write_init(PacketGameType.PACKET_CLIENT_GAME_INFO)
        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)



import enum
import logging

from ..wire.exceptions import PacketInvalidData
from ..wire.read import (
    read_string,
    read_uint8,
)
from ..wire.tcp import TCPProtocol

log = logging.getLogger(__name__)


class PacketStunType(enum.IntEnum):
    PACKET_STUN_SERCLI_STUN = 0
    PACKET_STUN_END = 1


class StunProtocol(TCPProtocol):
    PacketType = PacketStunType
    PACKET_END = PacketStunType.PACKET_STUN_END

    @staticmethod
    def receive_PACKET_STUN_SERCLI_STUN(source, data):
        protocol_version, data = read_uint8(data)

        if protocol_version < 3 or protocol_version > 6:
            raise PacketInvalidData("unknown protocol version: ", protocol_version)

        token, data = read_string(data)
        interface_number, data = read_uint8(data)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected; remaining: ", len(data))

        return {
            "protocol_version": protocol_version,
            "token": token,
            "interface_number": interface_number,
        }



import enum
import logging

from ..wire.exceptions import PacketInvalidData
from ..wire.read import (
    read_bytes,
    read_string,
    read_uint8,
    read_uint16,
    read_uint32,
)
from ..wire.tcp import TCPProtocol
from ..wire.write import (
    SEND_TCP_MTU,
    write_bytes,
    write_init,
    write_presend,
    write_string,
    write_uint8,
    write_uint16,
    write_uint32,
)

log = logging.getLogger(__name__)


# Value used to indicate no gamescript is loaded on the server.
# This is in fact (int32)-1 casted to an uint32.
GAMESCRIPT_VERSION_NONE = 4294967295
# The minimum starting year on the original TTD.
ORIGINAL_BASE_YEAR = 1920
# In GameInfo version 3 the date was changed to be counted from the year zero.
# This offset is added to version 2 and 1 to have the date the same for all
# versions. It is the amount of days from year 0 to 1920.
DAYS_TILL_ORIGINAL_BASE_YEAR = (
    365 * ORIGINAL_BASE_YEAR + ORIGINAL_BASE_YEAR // 4 - ORIGINAL_BASE_YEAR // 100 + ORIGINAL_BASE_YEAR // 400
)


class PacketCoordinatorType(enum.IntEnum):
    PACKET_COORDINATOR_GC_ERROR = 0
    PACKET_COORDINATOR_SERVER_REGISTER = 1
    PACKET_COORDINATOR_GC_REGISTER_ACK = 2
    PACKET_COORDINATOR_SERVER_UPDATE = 3
    PACKET_COORDINATOR_CLIENT_LISTING = 4
    PACKET_COORDINATOR_GC_LISTING = 5
    PACKET_COORDINATOR_CLIENT_CONNECT = 6
    PACKET_COORDINATOR_GC_CONNECTING = 7
    PACKET_COORDINATOR_SERCLI_CONNECT_FAILED = 8
    PACKET_COORDINATOR_GC_CONNECT_FAILED = 9
    PACKET_COORDINATOR_CLIENT_CONNECTED = 10
    PACKET_COORDINATOR_GC_DIRECT_CONNECT = 11
    PACKET_COORDINATOR_GC_STUN_REQUEST = 12
    PACKET_COORDINATOR_SERCLI_STUN_RESULT = 13
    PACKET_COORDINATOR_GC_STUN_CONNECT = 14
    PACKET_COORDINATOR_GC_NEWGRF_LOOKUP = 15
    PACKET_COORDINATOR_GC_TURN_CONNECT = 16
    PACKET_COORDINATOR_END = 17


class ServerGameType(enum.IntEnum):
    SERVER_GAME_TYPE_LOCAL = 0
    SERVER_GAME_TYPE_PUBLIC = 1
    SERVER_GAME_TYPE_INVITE_ONLY = 2
    SERVER_GAME_TYPE_END = 3


class ConnectionType(enum.IntEnum):
    CONNECTION_TYPE_UNKNOWN = 0
    CONNECTION_TYPE_ISOLATED = 1
    CONNECTION_TYPE_DIRECT = 2
    CONNECTION_TYPE_STUN = 3
    CONNECTION_TYPE_TURN = 4


class NetworkCoordinatorErrorType(enum.IntEnum):
    NETWORK_COORDINATOR_ERROR_UNKNOWN = 0
    NETWORK_COORDINATOR_ERROR_REGISTRATION_FAILED = 1
    NETWORK_COORDINATOR_ERROR_INVALID_INVITE_CODE = 2
    NETWORK_COORDINATOR_ERROR_REUSE_OF_INVITE_CODE = 3


class NewGRFSerializationType(enum.IntEnum):
    NST_GRFID_MD5 = 0
    NST_GRFID_MD5_NAME = 1
    NST_LOOKUP_ID = 2
    NST_END = 3
    # NST_CONVERSION_GRFID_MD5 is an internal value, assigned for those servers that didn't send this value yet.
    NST_CONVERSION_GRFID_MD5 = 4


class CoordinatorProtocol(TCPProtocol):
    PacketType = PacketCoordinatorType
    PACKET_END = PacketCoordinatorType.PACKET_COORDINATOR_END

    @staticmethod
    def receive_PACKET_COORDINATOR_SERVER_REGISTER(source, data):
        protocol_version, data = read_uint8(data)

        if protocol_version < 1 or protocol_version > 6:
            raise PacketInvalidData("unknown protocol version: ", protocol_version)

        game_type, data = read_uint8(data)

        if game_type >= ServerGameType.SERVER_GAME_TYPE_END:
            raise PacketInvalidData("invalid ServerGameType", game_type)

        game_type = ServerGameType(game_type)

        server_port, data = read_uint16(data)

        if protocol_version > 1:
            invite_code, data = read_string(data)
            invite_code_secret, data = read_string(data)
        else:
            invite_code = None
            invite_code_secret = None

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected in SERVER_REGISTER; remaining: ", len(data))

        return {
            "protocol_version": protocol_version,
            "game_type": game_type,
            "server_port": server_port,
            "invite_code": invite_code,
            "invite_code_secret": invite_code_secret,
        }

    @staticmethod
    def receive_PACKET_COORDINATOR_SERVER_UPDATE(source, data):
        protocol_version, data = read_uint8(data)

        if protocol_version < 1 or protocol_version > 6:
            raise PacketInvalidData("unknown protocol version: ", protocol_version)

        game_info_version, data = read_uint8(data)

        if game_info_version < 1 or game_info_version > 6:
            raise PacketInvalidData("unknown game info version: ", game_info_version)

        if game_info_version >= 6:
            newgrf_serialization_type, data = read_uint8(data)

            if newgrf_serialization_type >= NewGRFSerializationType.NST_END:
                raise PacketInvalidData("invalid NewGRFSerializationType", newgrf_serialization_type)

            newgrf_serialization_type = NewGRFSerializationType(newgrf_serialization_type)
            if newgrf_serialization_type == NewGRFSerializationType.NST_LOOKUP_ID:
                raise PacketInvalidData("NewGRF serialization type cannot be NST_LOOKUP_ID")
        else:
            newgrf_serialization_type = NewGRFSerializationType.NST_CONVERSION_GRFID_MD5

        if game_info_version >= 5:
            gamescript_version, data = read_uint32(data)
            gamescript_name, data = read_string(data)
        else:
            gamescript_version = None
            gamescript_name = None

        if game_info_version >= 4:
            newgrf_count, data = read_uint8(data)

            newgrfs = []
            for _ in range(newgrf_count):
                grfid, data = read_uint32(data)
                md5sum, data = read_bytes(data, 16)

                if newgrf_serialization_type == NewGRFSerializationType.NST_GRFID_MD5_NAME:
                    name, data = read_string(data)
                else:
                    name = None

                newgrfs.append({"grfid": grfid, "md5sum": md5sum.hex(), "name": name})
        else:
            newgrfs = None

        if game_info_version >= 3:
            game_date, data = read_uint32(data)
            start_date, data = read_uint32(data)

        if game_info_version >= 2:
            companies_max, data = read_uint8(data)
            companies_on, data = read_uint8(data)
            spectators_max, data = read_uint8(data)
        else:
            companies_max = None
            companies_on = None
            spectators_max = None

        if game_info_version >= 1:
            name, data = read_string(data)
            openttd_version, data = read_string(data)
            if game_info_version < 6:
                _, data = read_uint8(data)  # Unused, used to be server-lang
            use_password, data = read_uint8(data)

            clients_max, data = read_uint8(data)
            clients_on, data = read_uint8(data)
            spectators_on, data = read_uint8(data)

            if game_info_version < 3:
                game_date, data = read_uint16(data)
                game_date += DAYS_TILL_ORIGINAL_BASE_YEAR
                start_date, data = read_uint16(data)
                start_date += DAYS_TILL_ORIGINAL_BASE_YEAR

            if game_info_version < 6:
                _, data = read_string(data)  # Unused, used to be map-name
            map_width, data = read_uint16(data)
            map_height, data = read_uint16(data)
            map_type, data = read_uint8(data)

            is_dedicated, data = read_uint8(data)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected in SERVER_UPDATE; remaining: ", len(data))

        return {
            "protocol_version": protocol_version,
            "newgrf_serialization_type": newgrf_serialization_type,
            "newgrfs": newgrfs,
            "game_date": game_date,
            "start_date": start_date,
            "companies_max": companies_max,
            "companies_on": companies_on,
            "clients_max": clients_max,
            "clients_on": clients_on,
            "spectators_max": spectators_max,
            "spectators_on": spectators_on,
            "name": name,
            "openttd_version": openttd_version,
            "use_password": use_password,
            "is_dedicated": is_dedicated,
            "map_width": map_width,
            "map_height": map_height,
            "map_type": map_type,
            "gamescript_version": gamescript_version,
            "gamescript_name": gamescript_name,
        }

    @staticmethod
    def receive_PACKET_COORDINATOR_CLIENT_LISTING(source, data):
        protocol_version, data = read_uint8(data)

        if protocol_version < 1 or protocol_version > 6:
            raise PacketInvalidData("unknown protocol version: ", protocol_version)

        game_info_version, data = read_uint8(data)

        if game_info_version < 1 or game_info_version > 6:
            raise PacketInvalidData("unknown game info version: ", game_info_version)

        openttd_version, data = read_string(data)
        if protocol_version >= 4:
            newgrf_lookup_table_cursor, data = read_uint32(data)
        else:
            newgrf_lookup_table_cursor = None

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected in CLIENT_LISTING; remaining: ", len(data))

        return {
            "protocol_version": protocol_version,
            "game_info_version": game_info_version,
            "openttd_version": openttd_version,
            "newgrf_lookup_table_cursor": newgrf_lookup_table_cursor,
        }

    @staticmethod
    def receive_PACKET_COORDINATOR_CLIENT_CONNECT(source, data):
        protocol_version, data = read_uint8(data)

        if protocol_version < 2 or protocol_version > 6:
            raise PacketInvalidData("unknown protocol version: ", protocol_version)

        invite_code, data = read_string(data)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected in CLIENT_CONNECT; remaining: ", len(data))

        return {
            "protocol_version": protocol_version,
            "invite_code": invite_code,
        }

    @staticmethod
    def receive_PACKET_COORDINATOR_SERCLI_CONNECT_FAILED(source, data):
        protocol_version, data = read_uint8(data)

        if protocol_version < 2 or protocol_version > 6:
            raise PacketInvalidData("unknown protocol version: ", protocol_version)

        token, data = read_string(data)
        tracking_number, data = read_uint8(data)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected in SERCLI_CONNECT_FAILED; remaining: ", len(data))

        return {
            "protocol_version": protocol_version,
            "token": token,
            "tracking_number": tracking_number,
        }

    @staticmethod
    def receive_PACKET_COORDINATOR_CLIENT_CONNECTED(source, data):
        protocol_version, data = read_uint8(data)

        if protocol_version < 2 or protocol_version > 6:
            raise PacketInvalidData("unknown protocol version: ", protocol_version)

        token, data = read_string(data)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected in CLIENT_CONNECTED; remaining: ", len(data))

        return {
            "protocol_version": protocol_version,
            "token": token,
        }

    @staticmethod
    def receive_PACKET_COORDINATOR_SERCLI_STUN_RESULT(source, data):
        protocol_version, data = read_uint8(data)

        if protocol_version < 3 or protocol_version > 6:
            raise PacketInvalidData("unknown protocol version: ", protocol_version)

        token, data = read_string(data)
        interface_number, data = read_uint8(data)
        result, data = read_uint8(data)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected in SERCLI_STUN_RESULT; remaining: ", len(data))

        return {
            "protocol_version": protocol_version,
            "token": token,
            "interface_number": interface_number,
            "result": result,
        }

    async def send_PACKET_COORDINATOR_GC_ERROR(self, protocol_version, error_no, error_detail):
        data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_ERROR)

        # Older protocol versions didn't know "REUSE_OF_INVITE_CODE" yet. So
        # replace it with the next best thing: "REGISTRATION_FAILED".
        if (
            protocol_version < 6
            and error_no == NetworkCoordinatorErrorType.NETWORK_COORDINATOR_ERROR_REUSE_OF_INVITE_CODE
        ):
            error_no = NetworkCoordinatorErrorType.NETWORK_COORDINATOR_ERROR_REGISTRATION_FAILED

        write_uint8(data, error_no.value)
        write_string(data, error_detail)

        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)

    async def send_PACKET_COORDINATOR_GC_REGISTER_ACK(
        self, protocol_version, connection_type, invite_code, invite_code_secret
    ):
        data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_REGISTER_ACK)

        if protocol_version > 1:
            write_string(data, invite_code)
            write_string(data, invite_code_secret)
        write_uint8(data, connection_type.value)

        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)

    def _fill_NEWGRF_LOOKUP_PACKET(self, newgrf_lookup_table_cursor, newgrf_lookup_table):
        data = bytearray()
        count = 0
        for index, newgrf in newgrf_lookup_table.items():
            if index <= newgrf_lookup_table_cursor:
                continue

            count += 1

            write_uint32(data, index)
            write_uint32(data, newgrf["grfid"])
            write_bytes(data, bytes.fromhex(newgrf["md5sum"]))
            write_string(data, newgrf["name"] if newgrf["name"] is not None else "Unknown")

            # An entry is at most 4 + 4 + 16 + 80 = 104 bytes long. So use 200
            # as safe distance from SEND_TCP_MTU. Once reached, send the
            # packet and prepare for the next.
            if len(data) > SEND_TCP_MTU - 200:
                yield count, data
                data = bytearray()
                count = 0

        if count != 0:
            yield count, data

    async def send_PACKET_COORDINATOR_GC_NEWGRF_LOOKUP(
        self, protocol_version, newgrf_lookup_table_cursor, newgrf_lookup_table
    ):
        cursor = max(newgrf_lookup_table.keys())

        for count, body in self._fill_NEWGRF_LOOKUP_PACKET(newgrf_lookup_table_cursor, newgrf_lookup_table):
            data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_NEWGRF_LOOKUP)

            # The cursor is the highest index in the table. Index only increases
            # (till a full database reset), so it is a pretty safe cursor to use.
            write_uint32(data, cursor)
            write_uint16(data, count)
            write_bytes(data, body)

            write_presend(data, SEND_TCP_MTU)
            await self.send_packet(data)

    async def send_PACKET_COORDINATOR_GC_LISTING(
        self, protocol_version, game_info_version, servers, newgrf_lookup_table
    ):
        for server in servers:
            if server.game_type != ServerGameType.SERVER_GAME_TYPE_PUBLIC:
                continue
            if len(server.info) == 0:
                continue

            data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_LISTING)

            write_uint16(data, 1)

            write_string(data, server.connection_string)
            write_uint8(data, game_info_version)

            if game_info_version >= 6:
                write_uint8(data, NewGRFSerializationType.NST_LOOKUP_ID)

            if game_info_version >= 5:
                if server.info["gamescript_version"] is None or server.info["gamescript_name"] is None:
                    write_uint32(data, GAMESCRIPT_VERSION_NONE)
                    write_string(data, "")
                else:
                    write_uint32(data, server.info["gamescript_version"])
                    write_string(data, server.info["gamescript_name"])

            if game_info_version >= 4:
                write_uint8(data, len(server.newgrfs_indexed))
                if game_info_version >= 6:
                    for newgrf_indexed in server.newgrfs_indexed:
                        write_uint32(data, newgrf_indexed)
                else:
                    for newgrf_indexed in server.newgrfs_indexed:
                        newgrf = newgrf_lookup_table[newgrf_indexed]
                        write_uint32(data, newgrf["grfid"])
                        write_bytes(data, bytes.fromhex(newgrf["md5sum"]))

            if game_info_version >= 3:
                write_uint32(data, server.info["game_date"])
                write_uint32(data, server.info["start_date"])

            if game_info_version >= 2:
                write_uint8(data, server.info["companies_max"])
                write_uint8(data, server.info["companies_on"])
                write_uint8(data, server.info["spectators_max"])

            if game_info_version >= 1:
                write_string(data, server.info["name"])
                write_string(data, server.info["openttd_version"])
                if game_info_version <= 5:
                    write_uint8(data, 0)  # Unused, used to be server-lang
                write_uint8(data, server.info["use_password"])
                write_uint8(data, server.info["clients_max"])
                write_uint8(data, server.info["clients_on"])
                write_uint8(data, server.info["spectators_on"])

                if game_info_version < 3:
                    write_uint16(data, server.info["game_date"] - DAYS_TILL_ORIGINAL_BASE_YEAR)
                    write_uint16(data, server.info["start_date"] - DAYS_TILL_ORIGINAL_BASE_YEAR)

                if game_info_version <= 5:
                    write_string(data, "")  # Unused, used to be map-name
                write_uint16(data, server.info["map_width"])
                write_uint16(data, server.info["map_height"])
                write_uint8(data, server.info["map_type"])

                write_uint8(data, server.info["is_dedicated"])

            write_presend(data, SEND_TCP_MTU)
            await self.send_packet(data)

        # Send a final packet with 0 servers to indicate end-of-list.
        data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_LISTING)
        write_uint16(data, 0)
        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)

    async def send_PACKET_COORDINATOR_GC_CONNECTING(self, protocol_version, token, invite_code):
        data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_CONNECTING)

        write_string(data, token)
        write_string(data, invite_code)

        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)

    async def send_PACKET_COORDINATOR_GC_CONNECT_FAILED(self, protocol_version, token):
        data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_CONNECT_FAILED)

        write_string(data, token)

        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)

    async def send_PACKET_COORDINATOR_GC_DIRECT_CONNECT(self, protocol_version, token, tracking_number, hostname, port):
        data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_DIRECT_CONNECT)

        write_string(data, token)
        write_uint8(data, tracking_number)
        write_string(data, hostname)
        write_uint16(data, port)

        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)

    async def send_PACKET_COORDINATOR_GC_STUN_REQUEST(self, protocol_version, token):
        data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_STUN_REQUEST)

        write_string(data, token)

        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)

    async def send_PACKET_COORDINATOR_GC_STUN_CONNECT(
        self, protocol_version, token, tracking_number, interface_number, hostname, port
    ):
        data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_STUN_CONNECT)

        write_string(data, token)
        write_uint8(data, tracking_number)
        write_uint8(data, interface_number)
        write_string(data, hostname)
        write_uint16(data, port)

        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)

    async def send_PACKET_COORDINATOR_GC_TURN_CONNECT(
        self, protocol_version, token, tracking_number, ticket, connection_string
    ):
        data = write_init(PacketCoordinatorType.PACKET_COORDINATOR_GC_TURN_CONNECT)

        write_string(data, token)
        write_uint8(data, tracking_number)
        write_string(data, ticket)
        write_string(data, connection_string)

        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)



import struct

from .exceptions import PacketTooBig

SEND_TCP_MTU = 32767  # Since OpenTTD 1.12, OpenTTD clients support this MTU for TCP.
SEND_TCP_COMPAT_MTU = 1460  # Before OpenTTD 1.12, OpenTTD client support this MTU for TCP.
SEND_UDP_MTU = 1460  # OpenTTD clients support this MTU for UDP.


def write_uint8(data: bytearray, value: int) -> None:
    """Write a uint8 in the packet."""
    data += struct.pack("<B", value)


def write_uint16(data: bytearray, value: int) -> None:
    """Write a uint16 in the packet."""
    data += struct.pack("<H", value)


def write_uint32(data: bytearray, value: int) -> None:
    """Write a uint32 in the packet."""
    data += struct.pack("<I", value)


def write_uint64(data: bytearray, value: int) -> None:
    """Write a uint64 in the packet."""
    data += struct.pack("<Q", value)


def write_bytes(data: bytearray, value: bytes) -> None:
    """Write bytes in the packet."""
    data += value


def write_string(data: bytearray, value: str) -> None:
    """Write a string in the packet."""
    data += value.encode() + b"\x00"


def write_init(type: int) -> bytearray:
    """Initialize the writing of a new packet."""
    # write_presend() will replace this with the length of the packet.
    data = bytearray(b"\x00\x00")
    write_uint8(data, type)

    return data


def write_presend(data: bytearray, max_size: int) -> bytes:
    """Prepare a packet for sending. Returns an immutable list of bytes to send."""
    if len(data) > max_size:
        raise PacketTooBig(len(data))

    data[0:2] = struct.pack("<H", len(data))
    return bytes(data)



import ipaddress


class Source:
    def __init__(self, protocol, addr, ip, port):
        self.protocol = protocol
        self.addr = addr

        # Normally ip and port are in addr, but in case of Proxy Protocol
        # this might differ. So, please use ip/port over addr.
        self.ip = ipaddress.ip_address(ip)
        self.port = port

        # If using IPv6, IPv4 addresses are mapped like "::fffff:<IPv4>".
        # Convert those instances to an IPv4Address, so the class of an
        # instance can be used to easily detect if it is an IPv4 or IPv6.
        if isinstance(self.ip, ipaddress.IPv6Address) and self.ip.ipv4_mapped:
            self.ip = self.ip.ipv4_mapped

    def __repr__(self):
        return f"Source(ip={self.ip!r}, port={self.port!r}, addr={self.addr!r})"



import enum
import logging

from ..wire.exceptions import PacketInvalidData
from ..wire.read import (
    read_string,
    read_uint8,
)
from ..wire.tcp import TCPProtocol
from ..wire.write import (
    SEND_TCP_MTU,
    write_init,
    write_presend,
    write_string,
)

log = logging.getLogger(__name__)


class PacketTurnType(enum.IntEnum):
    PACKET_TURN_TURN_ERROR = 0
    PACKET_TURN_SERCLI_CONNECT = 1
    PACKET_TURN_TURN_CONNECTED = 2
    PACKET_TURN_END = 3


class TurnProtocol(TCPProtocol):
    PacketType = PacketTurnType
    PACKET_END = PacketTurnType.PACKET_TURN_END

    @staticmethod
    def receive_PACKET_TURN_SERCLI_CONNECT(source, data):
        protocol_version, data = read_uint8(data)

        if protocol_version < 5 or protocol_version > 6:
            raise PacketInvalidData("unknown protocol version: ", protocol_version)

        ticket, data = read_string(data)

        if len(data) != 0:
            raise PacketInvalidData("more bytes than expected; remaining: ", len(data))

        return {"protocol_version": protocol_version, "ticket": ticket}

    async def send_PACKET_TURN_TURN_CONNECTED(self, protocol_version, hostname):
        data = write_init(PacketTurnType.PACKET_TURN_TURN_CONNECTED)

        write_string(data, hostname)

        write_presend(data, SEND_TCP_MTU)
        await self.send_packet(data)
